
/*
 * $XConsortium$
 *
 * Copyright 1989 Massachusetts Institute of Technology
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of M.I.T. not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  M.I.T. makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * M.I.T. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL M.I.T.
 * BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Authors:  Davor Matic, MIT X Consortium
 *           Thomas Wu, MIT Fishbowl
 */

#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Intrinsic.h>

XImage *ProcessImage();
unsigned long DataSize();


typedef struct _Pair {
  Pixel pixel;
  long freq;
} Pair;

XImage *ProcessImage(dpy, scr, image, colormap, colors, ncolors, bc)
    Display *dpy;
    int scr;
    XImage *image;
    Colormap colormap;
    XColor *colors;
    Cardinal ncolors;
    Pixel *bc;
{
  Pixel *GetColorUsage(), *new_ranking;
  XImage *MapImage(), *new_image;
  Cardinal *pixel_freq;
  double ColorDistance();
  int i;

  if (image->depth > 1) {
      new_ranking = NULL;
      AllocateColors(dpy, scr, colormap, colors, ncolors, new_ranking);
      if(new_ranking)
	  free((char *) new_ranking);
      pixel_freq = (Cardinal *) XtCalloc((unsigned) ncolors, sizeof(Cardinal));
      new_image = MapImage(dpy, scr, image, colors, pixel_freq);
      *bc = 0;
      for (i = 1; i < ncolors; ++i)
	  if (pixel_freq[i] > pixel_freq[*bc])
	      *bc = i;
      *bc = colors[*bc].pixel;
      XDestroyImage(image);
      return new_image;
  }
  return image;
}

PCompare(p1, p2)
     Pair *p1, *p2;
{
  if(p1->freq < p2->freq)
    return 1;
  else if(p1->freq > p2->freq)
    return -1;
  else
    return 0;
}

Pixel *GetColorUsage(image, ncolors)
     XImage *image;
     Cardinal ncolors;
{
  Pair *usage_info;
  Pixel *rank_list;
  int i, j, unused;

  if((usage_info = (Pair *)XtMalloc((unsigned)ncolors * sizeof(Pair))) == NULL ||
     (rank_list = (Pixel *)XtMalloc((unsigned)ncolors * sizeof(Pixel))) == NULL) {
    XtWarning("Can't calloc color mapping storage.");
    exit(1);
  }
  for(i = 0; i < ncolors; ++i) {
    usage_info[i].pixel = i;
    usage_info[i].freq = 0;
  }
  for(i = 0; i < image->width; ++i)
    for(j = 0; j < image->height; ++j)
      ++usage_info[image->data[image->bytes_per_line * j + i]].freq;
  qsort((char *) usage_info, ncolors, sizeof(Pair), PCompare);
  unused = 0;
  for(i = 0; i < ncolors; ++i) {
    if(usage_info[i].freq == 0)
      ++unused;
    rank_list[i] = usage_info[i].pixel;
  }
  if(unused > 0)
    fprintf(stderr, "(%d colors unused.)\n", unused);
  return rank_list;
}

/*
 * Strategy for compromise:  If a color cannot be allocated (because the
 * colormap is full), find the closest existing color currently allocated.
 */
TryBestColorMatch(cl, test_color, num_colors)
XColor cl[], *test_color;
int num_colors;
{
  int i;
  long low_pixel_diff, pixel_diff, l_abs();

  low_pixel_diff = (-1);
  for(i = 0; i < num_colors; ++i) {
    pixel_diff = l_abs((long)(test_color->red - cl[i].red)) +
                 l_abs((long)(test_color->green - cl[i].green)) +
		 l_abs((long)(test_color->blue - cl[i].blue));
    if(low_pixel_diff < 0 || pixel_diff < low_pixel_diff) {
      test_color->pixel = cl[i].pixel;
      low_pixel_diff = pixel_diff;
    }
  }
}

long l_abs(x)
long x;
{
  return (x < 0) ? -x : x;
}

/*
 * Allocate all necessary colors in the default colormap, if possible.
 */
AllocateColors(dpy, screen, colormap, colors, ncolors, ranked_colors)
    Display *dpy;
    int screen;
    Colormap colormap;
    XColor *colors;
    Cardinal ncolors;
    Pixel *ranked_colors;
{
  Bool color_table_full = False;
  XColor *cmaplist;
  int i, j, index;
  int limit;

  limit = 1 << DefaultDepth(dpy, screen);
  cmaplist = (XColor *) XtMalloc(limit * sizeof(XColor));

  if(limit <= 0 || limit > ncolors)
    limit = ncolors;
  for(i = 0; i < ncolors; ++i) {
    index = ranked_colors ? ranked_colors[i] : i;
    if(i >= limit ||
       XAllocColor(dpy, colormap, &colors[index]) == 0) {
      if(!color_table_full && i < limit) {
	XtWarning("warning: cannot allocate all the desired colors.");
	for(j = 0; j < DisplayCells(dpy, screen); ++j)
	  cmaplist[j].pixel = j;
	XQueryColors(dpy, colormap, cmaplist, 
		     DisplayCells(dpy, screen));
	color_table_full = True;
      }
      TryBestColorMatch(cmaplist, &colors[index], DisplayCells(dpy, screen));
    }
  }
}

XImage *MapImage(dpy, scr, image, colors, pixel_freq)
    Display *dpy;
    int scr;
    XImage *image;
    XColor *colors;
    int *pixel_freq;
{
    Pixel pixel;
    XImage *new_image;
    int r, s;

    /*
     * Create image that matches default visual.
     */
    new_image = XCreateImage(dpy,
			     DefaultVisual(dpy, scr),
			     DefaultDepth(dpy, scr),
			     ZPixmap, image->xoffset,
			     XtMalloc(DataSize(image->bytes_per_line,
					       image->height, image->depth,
					       image->format)),
			     image->width, image->height,
			     image->bitmap_pad, image->bytes_per_line);

    for(r = 0; r < image->height; ++r)
	for(s = 0; s < image->width; ++s) {
	    ++pixel_freq[pixel = XGetPixel(image, s, r)];
	    XPutPixel(new_image, s, r, colors[pixel].pixel);
	}
    return new_image;
}

unsigned long DataSize(bpl, height, depth, format)
     int bpl, height, depth, format;
{
    if (format != ZPixmap)
	return (bpl * height * depth);
    else
	return(bpl * height);
}
